@@ -74,9 +74,18 @@ g_open (const gchar *filename,
   if (G_WIN32_HAVE_WIDECHAR_API ())
     {
       wchar_t *wfilename = g_utf8_to_utf16 (filename, -1, NULL, NULL, NULL);
-      int retval = _wopen (wfilename, flags, mode);
-      int save_errno = errno;
+      int retval;
+      int save_errno;
       
+      if (wfilename == NULL)
+	{
+	  errno = EINVAL;
+	  return -1;
+	}
+
+      retval = _wopen (wfilename, flags, mode);
+      save_errno = errno;
+
       g_free (wfilename);
 
       errno = save_errno;
@@ -85,8 +94,17 @@ g_open (const gchar *filename,
   else
     {    
       gchar *cp_filename = g_locale_from_utf8 (filename, -1, NULL, NULL, NULL);
-      int retval = open (cp_filename, flags, mode);
-      int save_errno = errno;
+      int retval;
+      int save_errno;
+
+      if (cp_filename == NULL)
+	{
+	  errno = EINVAL;
+	  return -1;
+	}
+
+      retval = open (cp_filename, flags, mode);
+      save_errno = errno;
 
       g_free (cp_filename);
 
@@ -106,7 +124,11 @@ g_open (const gchar *filename,
  * A wrapper for the POSIX rename() function. The rename() function 
  * renames a file, moving it between directories if required.
  * 
- * See the C library manual for more details about rename().
+ * See your C library manual for more details about how rename() works
+ * on your system. Note in particular that on Windows, it is in
+ * general not possible to rename a file if a file with the new name
+ * already exists. Also it is not possible in general to rename an
+ * open file.
  *
  * Returns: 0 if the renaming succeeded, -1 if an error occurred
  * 
@@ -120,10 +142,28 @@ g_rename (const gchar *oldfilename,
   if (G_WIN32_HAVE_WIDECHAR_API ())
     {
       wchar_t *woldfilename = g_utf8_to_utf16 (oldfilename, -1, NULL, NULL, NULL);
-      wchar_t *wnewfilename = g_utf8_to_utf16 (newfilename, -1, NULL, NULL, NULL);
-      int retval = _wrename (woldfilename, wnewfilename);
-      int save_errno = errno;
-      
+      wchar_t *wnewfilename;
+      int retval;
+      int save_errno;
+
+      if (woldfilename == NULL)
+	{
+	  errno = EINVAL;
+	  return -1;
+	}
+
+      wnewfilename = g_utf8_to_utf16 (newfilename, -1, NULL, NULL, NULL);
+
+      if (wnewfilename == NULL)
+	{
+	  g_free (woldfilename);
+	  errno = EINVAL;
+	  return -1;
+	}
+
+      retval = _wrename (woldfilename, wnewfilename);
+      save_errno = errno;
+
       g_free (woldfilename);
       g_free (wnewfilename);
       
@@ -133,9 +173,27 @@ g_rename (const gchar *oldfilename,
   else
     {
       gchar *cp_oldfilename = g_locale_from_utf8 (oldfilename, -1, NULL, NULL, NULL);
-      gchar *cp_newfilename = g_locale_from_utf8 (newfilename, -1, NULL, NULL, NULL);
-      int retval = rename (cp_oldfilename, cp_newfilename);
-      int save_errno = errno;
+      gchar *cp_newfilename;
+      int retval;
+      int save_errno;
+
+      if (cp_oldfilename == NULL)
+	{
+	  errno = EINVAL;
+	  return -1;
+	}
+
+      cp_newfilename = g_locale_from_utf8 (newfilename, -1, NULL, NULL, NULL);
+
+      if (cp_newfilename == NULL)
+	{
+	  g_free (cp_oldfilename);
+	  errno = EINVAL;
+	  return -1;
+	}
+	
+      retval = rename (cp_oldfilename, cp_newfilename);
+      save_errno = errno;
 
       g_free (cp_oldfilename);
       g_free (cp_newfilename);
@@ -171,8 +229,17 @@ g_mkdir (const gchar *filename,
   if (G_WIN32_HAVE_WIDECHAR_API ())
     {
       wchar_t *wfilename = g_utf8_to_utf16 (filename, -1, NULL, NULL, NULL);
-      int retval = _wmkdir (wfilename);
-      int save_errno = errno;
+      int retval;
+      int save_errno;
+
+      if (wfilename == NULL)
+	{
+	  errno = EINVAL;
+	  return -1;
+	}
+
+      retval = _wmkdir (wfilename);
+      save_errno = errno;
 
       g_free (wfilename);
       
@@ -182,8 +249,17 @@ g_mkdir (const gchar *filename,
   else
     {
       gchar *cp_filename = g_locale_from_utf8 (filename, -1, NULL, NULL, NULL);
-      int retval = mkdir (cp_filename);
-      int save_errno = errno;
+      int retval;
+      int save_errno;
+
+      if (cp_filename == NULL)
+	{
+	  errno = EINVAL;
+	  return -1;
+	}
+
+      retval = mkdir (cp_filename);
+      save_errno = errno;
 
       g_free (cp_filename);
 
@@ -219,8 +295,17 @@ g_stat (const gchar *filename,
   if (G_WIN32_HAVE_WIDECHAR_API ())
     {
       wchar_t *wfilename = g_utf8_to_utf16 (filename, -1, NULL, NULL, NULL);
-      int retval = _wstat (wfilename, (struct _stat *) buf);
-      int save_errno = errno;
+      int retval;
+      int save_errno;
+
+      if (wfilename == NULL)
+	{
+	  errno = EINVAL;
+	  return -1;
+	}
+
+      retval = _wstat (wfilename, (struct _stat *) buf);
+      save_errno = errno;
 
       g_free (wfilename);
 
@@ -230,8 +315,17 @@ g_stat (const gchar *filename,
   else
     {
       gchar *cp_filename = g_locale_from_utf8 (filename, -1, NULL, NULL, NULL);
-      int retval = stat (cp_filename, buf);
-      int save_errno = errno;
+      int retval;
+      int save_errno;
+
+      if (cp_filename == NULL)
+	{
+	  errno = EINVAL;
+	  return -1;
+	}
+
+      retval = stat (cp_filename, buf);
+      save_errno = errno;
 
       g_free (cp_filename);
 
@@ -283,9 +377,11 @@ g_lstat (const gchar *filename,
  * file and no processes have it opened, the diskspace occupied by the
  * file is freed.
  * 
- * See the C library manual for more details about unlink().
+ * See your C library manual for more details about unlink(). Note
+ * that on Windows, it is in general not possible to delete files that
+ * are open to some process, or mapped into memory.
  *
- * Returns: 0 if the directory was successfully created, -1 if an error 
+ * Returns: 0 if the name was successfully deleted, -1 if an error 
  *    occurred
  * 
  * Since: 2.6
@@ -297,8 +393,17 @@ g_unlink (const gchar *filename)
   if (G_WIN32_HAVE_WIDECHAR_API ())
     {
       wchar_t *wfilename = g_utf8_to_utf16 (filename, -1, NULL, NULL, NULL);
-      int retval = _wunlink (wfilename);
-      int save_errno = errno;
+      int retval;
+      int save_errno;
+
+      if (wfilename == NULL)
+	{
+	  errno = EINVAL;
+	  return -1;
+	}
+
+      retval = _wunlink (wfilename);
+      save_errno = errno;
 
       g_free (wfilename);
 
@@ -308,8 +413,17 @@ g_unlink (const gchar *filename)
   else
     {
       gchar *cp_filename = g_locale_from_utf8 (filename, -1, NULL, NULL, NULL);
-      int retval = unlink (cp_filename);
-      int save_errno = errno;
+      int retval;
+      int save_errno;
+
+      if (cp_filename == NULL)
+	{
+	  errno = EINVAL;
+	  return -1;
+	}
+
+      retval = unlink (cp_filename);
+      save_errno = errno;
 
       g_free (cp_filename);
 
@@ -325,13 +439,19 @@ g_unlink (const gchar *filename)
  * g_remove:
  * @filename: a pathname in the GLib file name encoding
  *
- * A wrapper for the POSIX remove() function. The remove() function 
- * deletes a name from the filesystem. It calls unlink() for files
- * and rmdir() for directories.
+ * A wrapper for the POSIX remove() function. The remove() function
+ * deletes a name from the filesystem.
  * 
- * See the C library manual for more details about remove().
+ * See your C library manual for more details about how remove() works
+ * on your system. On Unix, remove() removes also directories, as it
+ * calls unlink() for files and rmdir() for directories. On Windows,
+ * although remove() in the C library only works for files, this
+ * function tries first remove() and then if that fails rmdir(), and
+ * thus works for both files and directories. Note however, that on
+ * Windows, it is in general not possible to remove a file that is
+ * open to some process, or mapped into memory.
  *
- * Returns: 0 if the directory was successfully created, -1 if an error 
+ * Returns: 0 if the file was successfully removed, -1 if an error 
  *    occurred
  * 
  * Since: 2.6
@@ -343,8 +463,19 @@ g_remove (const gchar *filename)
   if (G_WIN32_HAVE_WIDECHAR_API ())
     {
       wchar_t *wfilename = g_utf8_to_utf16 (filename, -1, NULL, NULL, NULL);
-      int retval = _wremove (wfilename);
-      int save_errno = errno;
+      int retval;
+      int save_errno;
+
+      if (wfilename == NULL)
+	{
+	  errno = EINVAL;
+	  return -1;
+	}
+
+      retval = _wremove (wfilename);
+      if (retval == -1)
+	retval = _wrmdir (wfilename);
+      save_errno = errno;
 
       g_free (wfilename);
 
@@ -354,8 +485,19 @@ g_remove (const gchar *filename)
   else
     {
       gchar *cp_filename = g_locale_from_utf8 (filename, -1, NULL, NULL, NULL);
-      int retval = remove (cp_filename);
-      int save_errno = errno;
+      int retval;
+      int save_errno;
+      
+      if (cp_filename == NULL)
+	{
+	  errno = EINVAL;
+	  return -1;
+	}
+
+      retval = remove (cp_filename);
+      if (retval == -1)
+	retval = rmdir (cp_filename);
+      save_errno = errno;
 
       g_free (cp_filename);
 
@@ -367,6 +509,70 @@ g_remove (const gchar *filename)
 #endif
 }
 
+/**
+ * g_rmdir:
+ * @filename: a pathname in the GLib file name encoding
+ *
+ * A wrapper for the POSIX rmdir() function. The rmdir() function
+ * deletes a directory from the filesystem.
+ * 
+ * See your C library manual for more details about how rmdir() works
+ * on your system.
+ *
+ * Returns: 0 if the directory was successfully removed, -1 if an error 
+ *    occurred
+ * 
+ * Since: 2.6
+ */
+int
+g_rmdir (const gchar *filename)
+{
+#ifdef G_OS_WIN32
+  if (G_WIN32_HAVE_WIDECHAR_API ())
+    {
+      wchar_t *wfilename = g_utf8_to_utf16 (filename, -1, NULL, NULL, NULL);
+      int retval;
+      int save_errno;
+
+      if (wfilename == NULL)
+	{
+	  errno = EINVAL;
+	  return -1;
+	}
+      
+      retval = _wrmdir (wfilename);
+      save_errno = errno;
+
+      g_free (wfilename);
+
+      errno = save_errno;
+      return retval;
+    }
+  else
+    {
+      gchar *cp_filename = g_locale_from_utf8 (filename, -1, NULL, NULL, NULL);
+      int retval;
+      int save_errno;
+
+      if (cp_filename == NULL)
+	{
+	  errno = EINVAL;
+	  return -1;
+	}
+
+      retval = rmdir (cp_filename);
+      save_errno = errno;
+
+      g_free (cp_filename);
+
+      errno = save_errno;
+      return retval;
+    }
+#else
+  return rmdir (filename);
+#endif
+}
+
 /**
  * g_fopen:
  * @filename: a pathname in the GLib file name encoding
@@ -391,9 +597,27 @@ g_fopen (const gchar *filename,
   if (G_WIN32_HAVE_WIDECHAR_API ())
     {
       wchar_t *wfilename = g_utf8_to_utf16 (filename, -1, NULL, NULL, NULL);
-      wchar_t *wmode = g_utf8_to_utf16 (mode, -1, NULL, NULL, NULL);
-      FILE *retval = _wfopen (wfilename, wmode);
-      int save_errno = errno;
+      wchar_t *wmode;
+      FILE *retval;
+      int save_errno;
+
+      if (wfilename == NULL)
+	{
+	  errno = EINVAL;
+	  return NULL;
+	}
+
+      wmode = g_utf8_to_utf16 (mode, -1, NULL, NULL, NULL);
+
+      if (wmode == NULL)
+	{
+	  g_free (wfilename);
+	  errno = EINVAL;
+	  return NULL;
+	}
+	
+      retval = _wfopen (wfilename, wmode);
+      save_errno = errno;
 
       g_free (wfilename);
       g_free (wmode);
@@ -404,8 +628,17 @@ g_fopen (const gchar *filename,
   else
     {
       gchar *cp_filename = g_locale_from_utf8 (filename, -1, NULL, NULL, NULL);
-      FILE *retval = fopen (cp_filename, mode);
-      int save_errno = errno;
+      FILE *retval;
+      int save_errno;
+
+      if (cp_filename == NULL)
+	{
+	  errno = EINVAL;
+	  return NULL;
+	}
+
+      retval = fopen (cp_filename, mode);
+      save_errno = errno;
 
       g_free (cp_filename);
 
@@ -443,9 +676,27 @@ g_freopen (const gchar *filename,
   if (G_WIN32_HAVE_WIDECHAR_API ())
     {
       wchar_t *wfilename = g_utf8_to_utf16 (filename, -1, NULL, NULL, NULL);
-      wchar_t *wmode = g_utf8_to_utf16 (mode, -1, NULL, NULL, NULL);
-      FILE *retval = _wfreopen (wfilename, wmode, stream);
-      int save_errno = errno;
+      wchar_t *wmode;
+      FILE *retval;
+      int save_errno;
+
+      if (wfilename == NULL)
+	{
+	  errno = EINVAL;
+	  return NULL;
+	}
+      
+      wmode = g_utf8_to_utf16 (mode, -1, NULL, NULL, NULL);
+
+      if (wmode == NULL)
+	{
+	  g_free (wfilename);
+	  errno = EINVAL;
+	  return NULL;
+	}
+      
+      retval = _wfreopen (wfilename, wmode, stream);
+      save_errno = errno;
 
       g_free (wfilename);
       g_free (wmode);
@@ -456,8 +707,17 @@ g_freopen (const gchar *filename,
   else
     {
       gchar *cp_filename = g_locale_from_utf8 (filename, -1, NULL, NULL, NULL);
-      FILE *retval = freopen (cp_filename, mode, stream);
-      int save_errno = errno;
+      FILE *retval;
+      int save_errno;
+
+      if (cp_filename == NULL)
+	{
+	  errno = EINVAL;
+	  return NULL;
+	}
+
+      retval = freopen (cp_filename, mode, stream);
+      save_errno = errno;
 
       g_free (cp_filename);
 
