@@ -31,7 +31,8 @@
  * what they are doing and no exchange of files with others this would
  * be a well-defined encoding, preferrably UTF-8. On Windows, the
  * pathname is always in UTF-8, even if that is not the on-disk
- * encoding or the encoding used by the Win32 API.
+ * encoding, and not the encoding accepted by the C library or Win32
+ * API.
  */
 
 int g_open      (const gchar *filename,
@@ -54,6 +55,8 @@ int g_unlink    (const gchar *filename);
 
 int g_remove    (const gchar *filename);
 
+int g_rmdir (const gchar *filename);
+
 FILE *g_fopen   (const gchar *filename,
                  const gchar *mode);
 
